#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug 23 20:35:43 2022

@author: frem
"""

from ekde import KDE
import numpy as np
from matplotlib import pyplot as plt

x_train = np.random.normal(0, 1, 10**3)

kde = KDE(kernel='gaussian').fit(x_train[:,None])
x_test = np.linspace(-5,5,100)
y_test = kde.predict(x_test[:,None])

plt.plot(x_test, y_test)

print(kde.predict(np.array([1,1,1,1,0,0,0])[:,None]))


# distutils: language = c++

from libcpp cimport bool
from libc.stdlib cimport malloc, free
import cython

from libc.math cimport exp
from libc.math cimport pow as cpow

from tqdm import tqdm

import numpy as np
cimport numpy as np

from cpython cimport array
import array     

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cpdef void count_diff_desc(int[:,:] A,
               int[:,:] out):
    cdef Py_ssize_t j, i_asc, i
    
    cdef int n = A.shape[0]
    cdef int d = A.shape[1]
    
    for i_asc in range(n - 1):
        i = n - 2 - i_asc
        for j in range(d):
            if A[i,j] == A[i+1, j]:
                if j == 0:
                    out[i,j] = out[i+1,j] + 1
                elif out[i,j-1] > 1:
                    out[i,j] = out[i+1,j] + 1  

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cpdef void count_diff_asc(int[:,:] A,
                          int[:,:] out):
    cdef Py_ssize_t j, i
    
    cdef int n = A.shape[0]
    cdef int d = A.shape[1]
    
    for i in range(1, n):
        for j in range(d):
            if A[i,j] == A[i-1, j]:
                if j == 0:
                    out[i,j] = out[i-1,j] + 1
                elif out[i,j-1] > 1:
                    out[i,j] = out[i-1,j] + 1


@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef int *** sparse(int[:,:] A,
                    int[:,:] A_diff_desc,
                    int[:,:] A_diff_one_side,
                    int *S_shape):
    cdef Py_ssize_t i_A, i_S, j, j_asc
    
    cdef int n = A.shape[0]
    cdef int d = A.shape[1]
        
    cdef int ***S = <int ***> malloc(d *sizeof(int **))
    
    for j in range(d):
        S[j] = <int **> malloc(S_shape[j] * sizeof(int *))
    
    for i_A in range(n):
        S[d-1][i_A] = <int *> malloc(2 * sizeof(int))
        S[d-1][i_A][0] = A[i_A, d-1]
        S[d-1][i_A][1] = i_A
    
    for j in range(d-1):
        S[j][0] = <int *> malloc(2 * sizeof(int))
        S[j][0][0] = A[0, j]
        S[j][0][1] = 0
    
    for j in range(d-1):
        i_A = 0
        for i_S in range(1, S_shape[j]):
            S[j][i_S] = <int *> malloc(2 * sizeof(int))
            
            S[j][i_S][1] = S[j][i_S - 1][1] + A_diff_one_side[i_A, j]
            
            i_A = i_A + A_diff_desc[i_A, j]
            
            S[j][i_S][0] = A[i_A, j]                
    
    return(S)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef int [:,:] count_one_side(int[:,:] A_diff_desc):
    cdef Py_ssize_t i, j, i_asc
    
    cdef int n = A_diff_desc.shape[0]
    cdef int d = A_diff_desc.shape[1]
    cdef int [:,:] A_diff_one_side = np.zeros((n, d-1), dtype=np.intc)
    
    cdef int cnt
    
    for j in range(d-1):
        cnt = 0
        for i_asc in range(n):
            i = n - 1 - i_asc
            if A_diff_desc[i, j] == 1:
                cnt = 0
                
            if A_diff_desc[i, j+1] == 1:
                cnt = cnt + 1
            
            A_diff_one_side[i, j] = cnt
    
    return(A_diff_one_side)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef int exponential_search_left(int **L,
                                 int x,
                                 int a,
                                 int b):
    cdef int bound = 1
    cdef size = b - a
    
    while bound < size and L[a + bound][0] < x:
        bound = bound * 2
        
    return(binary_search_left(L=L, x=x, a=a + bound/2, b=min(a + bound + 1, b)))

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef int exponential_search_right(int **L,
                                  int x,
                                  int a,
                                  int b):
    cdef int bound = 1
    cdef size = b - a
    
    while bound < size and L[a + bound][0] <= x:
        bound = bound * 2
    
    return(binary_search_right(L=L, x=x, a=a + bound/2, b=min(a + bound + 1, b)))

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef int binary_search_left(int **L,
                            int x,
                            int a,
                            int b):
    cdef int m
    
    if a >= b:
        return(a)
    
    if x < L[a][0]:
        return(a)
    
    if x > L[b-1][0]:
        return(b)
    
    while a < b:
        m = <int> (a + b) / 2
        
        if L[m][0] < x:
            a = m + 1
        else:
            b = m
    
    return(a)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef int binary_search_right(int **L,
                             int x,
                             int a,
                             int b):
    cdef int m
    
    if b <= a:
        return(b)
    
    if x < L[a][0]:
        return(a)
    
    if x > L[b-1][0]:
        return(b)
    
    while a < b:
        m = <int> (a + b) / 2
        
        if L[m][0] > x:
            b = m
        else:
            a = m +1
    
    return(b)

        
@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef int * get_S_shape(int[:,:] U_diff_desc, 
                       int n_U,
                       int d):
    cdef Py_ssize_t j
    
    cdef int *S_shape = <int *> malloc(d * sizeof(int))
    
    for j in range(d):
        S_shape[j] = 0
        
        for i_U in range(n_U):
            if U_diff_desc[i_U, j] == 1:
                S_shape[j] = S_shape[j] + 1
    return(S_shape)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cpdef double [:] set_estimation(int[:,:] Z_diff_asc,
                                int[:] Z_indices,
                                double[:] g):
    cdef Py_ssize_t i_Z, i_g
    cdef int n_Z = Z_diff_asc.shape[0]
    cdef int d = Z_diff_asc.shape[1]
    
    cdef double [:] f = np.zeros(n_Z, dtype=np.double)
    
    i_g = -1
    for i_Z in range(n_Z):
        if Z_diff_asc[i_Z, d-1] == 1:
            i_g = i_g + 1
        f[Z_indices[i_Z]] = g[i_g]
    return(f)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)  # non check division 
def merge(int[:, :] U, 
          int[:, :] U_diff_desc,
          double[:] nu, 
          int[:, :] Z,
          int q,
          double h,
          int kernel_id,
          double dx,
          int verbose=0):
    cdef Py_ssize_t i_U, i_Z, j, k, i_T
    
    cdef int n_U = U.shape[0]
    cdef int d = U.shape[1]
    cdef int n_Z = Z.shape[0]
    
    # get the margin i.e the kernel support radius 
    # in cell unit around a kernel center.
    cdef int margin = (q - 1) / 2
    
    # initialize the output array
    cdef double [:] f = np.zeros(n_Z, dtype=np.double)
    
    # U sparsing -> S
    cdef int [:,:] U_diff_one_side = count_one_side(A_diff_desc=U_diff_desc)
    cdef int *S_shape = get_S_shape(U_diff_desc,
                                    n_U,
                                    d)
    cdef int ***S = sparse(A=U,
                           A_diff_desc=U_diff_desc,
                           A_diff_one_side = U_diff_one_side,
                           S_shape=S_shape)
    
    # Z sparsing -> T
    cdef int[:,:] Z_diff_desc = np.ones((n_Z, d), dtype=np.intc)
    count_diff_desc(Z, Z_diff_desc)
    cdef int [:,:] Z_diff_one_side = count_one_side(A_diff_desc=Z_diff_desc)
    cdef int *T_shape = get_S_shape(Z_diff_desc,
                                    n_Z,
                                    d)
    cdef int ***T = sparse(A=Z,
                           A_diff_desc=Z_diff_desc,
                           A_diff_one_side = Z_diff_one_side,
                           S_shape=T_shape)
    
    
    # initialize the exploration of the first column (j = 0)
    cdef int low_S = 0
    cdef int high_S = S_shape[0]
    
    # verbose initialization (tqdm package)
    if verbose > 0:
        for_list = tqdm(range(0, T_shape[0]))
    else:
        for_list = range(0, T_shape[0])
    
    # exploration of the first column
    # it is a reccursive loop to explore each columns
    for i_T in for_list:
        # the low_S is returned to reduce the S list length
        # at each loop (because T is increasing).
        
        low_S = explore(S=S,
                        S_shape=S_shape,
                        nu = nu,
                        d=d,
                        T = T,
                        T_shape=T_shape,
                        f=f,
                        margin=margin,
                        j=0,
                        low_S=0,
                        high_S=high_S,
                        i_T=i_T,
                        dx=dx,
                        h=h,
                        kernel_id=kernel_id,
                        dist_sq = 0)
    
    # free allocated memory
    for j in range(d):
        for i_U in range(S_shape[j]):
            free(S[j][i_U])
        
        for i_T in range(T_shape[j]):
            free(T[j][i_T])
        
        free(S[j])
        free(T[j])
    free(S)
    free(T)
    
    free(S_shape)
    free(T_shape)
    
    return(np.array(f))

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)  # non check division 
cdef int explore(int ***S, 
                 int *S_shape, 
                 double[:] nu,
                 int d,
                 int ***T,
                 int *T_shape,
                 double[:] f,
                 int margin,
                 int j,
                 int low_S,
                 int high_S,
                 int i_T,
                 double dx,
                 double h, 
                 int kernel_id,
                 double dist_sq):
    cdef Py_ssize_t i_S
    
    cdef int z = T[j][i_T][0]
    cdef int low_T = T[j][i_T][1]
    cdef int high_T = get_high(T[j], T_shape, d, j, i_T)
    cdef int i_Z
    cdef int a, b
    
    a = binary_search_left(L = S[j],
                            x = z - margin,
                            a = low_S,
                            b = high_S)
    b = binary_search_right(L = S[j],
                            x = z + margin,
                            a = a,
                            b = high_S)
    # a = exponential_search_left(L = S[j],
    #                             x = z - margin,
    #                             a = low_S,
    #                             b = high_S)
    # b = exponential_search_right(L = S[j],
    #                              x = z + margin,
    #                              a = a,
    #                              b = high_S)
    
    cdef int low_S_prime = a
    
    if j < d-1:
        for i_S in range(a, b):
            low_S = S[j][i_S][1]
            high_S = get_high(S[j], S_shape, d, j, i_S)
            
            if kernel_id > 0:
                dist_sq = dist_sq + cpow(z - S[j][i_S][0], 2.0)
            
            for i_T in range(low_T, high_T):
                
                    
                low_S = explore(S=S,
                                S_shape=S_shape,
                                nu = nu,
                                d=d,
                                T=T,
                                T_shape=T_shape,
                                f=f,
                                margin=margin,
                                j=j+1,
                                low_S=low_S,
                                high_S=high_S,
                                i_T = i_T,
                                dx = dx,
                                h = h,
                                kernel_id = kernel_id,
                                dist_sq = dist_sq)
    else:
        i_Z = T[j][i_T][1]
        
        if kernel_id == 0:
            # box kernel
            for i_S in range(a, b):
                f[i_Z] = f[i_Z] + nu[S[j][i_S][1]]
                
        elif kernel_id == 1:
            # gaussian kernel
            for i_S in range(a, b):
                dist_sq = ( dist_sq + cpow(z - S[j][i_S][0], 2.0) ) * cpow(dx, 2.0)
                
                f[i_Z] = f[i_Z] + nu[S[j][i_S][1]] * exp(-dist_sq / cpow(h,2.0) / 2)
        
        elif kernel_id == 2:
            for i_S in range(a, b):
                dist_sq = ( dist_sq + cpow(z - S[j][i_S][0], 2.0) ) * cpow(dx, 2.0)
                
                if dist_sq < 1:
                    f[i_Z] = f[i_Z] + nu[S[j][i_S][1]] * 0.75 * (1 - dist_sq) ** 2
                    
    
    return(low_S_prime)


@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef get_high(int** S_j,
              int* S_shape,
              int d,
              int j,
              int i):
    
    if i == S_shape[j] - 1:
        if j == d - 1:
            return(S_j[i][1] + 1)
        else:
            return(S_shape[j+1])
    else:
        return(S_j[i+1][1])
    
@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)  # non check division 
cpdef double[:] boundary_correction_gaussian_kernel(double[:,:] A,
                                          double[:] R,
                                          double[:,:] C,
                                          double h,
                                          int n_mc=10000,
                                          double zero=0.000000000001):
    cdef Py_ssize_t i, i_hyp, j, i_mc
    
    cdef int d = A.shape[0]
    cdef int n_hyp = A.shape[1]
    cdef int n = C.shape[0]
    
    cdef double normalization = (2 * np.pi)**(d/2) * h**d
    
    cdef double [:] norm_2_a_square = np.linalg.norm(A, axis=0, ord=2)**2
    cdef double [:] norm_inf_a = np.linalg.norm(A, axis=0, ord=np.inf)
    
    cdef double[:,:] X_mc = np.random.random((n_mc, d))
    
    cdef double[:] g = np.zeros(n_mc, dtype=np.double)
    
    cdef double dist_sq
    for i_mc in range(n_mc):
        dist_sq = 0.0
        for j in range(d):
            dist_sq = dist_sq + cpow(X_mc[i_mc,j] - 0.5, 2.0)
        
        g[i_mc] = exp(-dist_sq / cpow(h,2.0) / 2) / normalization
        
        
    cdef double[:] V = np.ones(n, dtype=np.double)
    
    cdef double dist
    
    cdef double dist_max = h * 3.0
        
    cdef double [:,:] A_x = np.zeros((d, n_hyp), dtype=np.double)
    cdef double [:] R_x = np.zeros(n_hyp, dtype = np.double)
    cdef int m_x
    
    cdef double s
    
    for i in range(n):
        m_x = 0
        # distances to hyperplanes
        for i_hyp in range(n_hyp):
            dist = R[i_hyp]
            for j in range(d):
                dist = dist + C[i, j] * A[j, i_hyp]
            dist = abs(dist) / norm_2_a_square[i_hyp] * norm_inf_a[i_hyp]
            
            if dist <= dist_max:
                for j in range(d):
                    A_x[j, m_x] = A[j, i_hyp]
                R_x[m_x] = R[i_hyp]
                
                # affine transformation
                for j in range(d):
                    R_x[m_x] = R_x[m_x] - A_x[j, m_x] * (-C[i,j] + 3 * h)
                    A_x[j, m_x] = A_x[j, m_x] * 6 * h
                
                m_x = m_x + 1
        
        
        if m_x > 0:
            # count side
            s = 0
            for i_mc in range(n_mc):
                
                for i_hyp in range(m_x):
                    dist = zero + R_x[i_hyp]
                    for j in range(d):
                        dist = dist + X_mc[i_mc, j] * A_x[j, i_hyp]
                    
                    if dist <= 0:
                        break
                else:
                    s = s + g[i_mc]
            
            V[i] = s / n_mc
            
            if V[i] <= zero:
                V[i] = 1.0
                
    return(V)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 20 12:53:42 2022

@author: frem
"""

# this example is based on the ZCA the ZCA-Mahalanobis whitening transformation procedure (ZCA stands for ``zero-phase components analysis''). See \Kessy et al 2018 for more details.

# @Article{Kessy2018,
#   author    = {Kessy, A. and Lewin, A. and Strimmer, K.},
#   journal   = {The American Statistician},
#   title     = {Optimal {Whitening} and {Decorrelation}},
#   year      = {2018},
#   IGNOREissn      = {0003-1305},
#   IGNOREmonth     = oct,
#   number    = {4},
#   pages     = {309--314},
#   volume    = {72},
#   IGNOREabstract  = {Whitening, or sphering, is a common preprocessing step in statistical analysis to transform random variables to orthogonality. However, due to rotational freedom there are infinitely many possible whitening procedures. Consequently, there is a diverse range of sphering methods in use, for example, based on principal component analysis (PCA), Cholesky matrix decomposition, and zero-phase component analysis (ZCA), among others. Here, we provide an overview of the underlying theory and discuss five natural whitening procedures. Subsequently, we demonstrate that investigating the cross-covariance and the cross-correlation matrix between sphered and original variables allows to break the rotational invariance and to identify optimal whitening transformations. As a result we recommend two particular approaches: ZCA-cor whitening to produce sphered variables that are maximally similar to the original variables, and PCA-cor whitening to obtain sphered variables that maximally compress the original variables.},
#   file      = {:storage/LUCC/Kessy2018.pdf:PDF},
#   groups    = {Whitening Transformation},
#   IGNOREkeywords  = {CAR score, CAT score, Cholesky decomposition, Decorrelation, Principal components analysis, Whitening, ZCA-Mahalanobis transformation},
#   IGNOREpublisher = {Taylor \& Francis},
#   IGNOREurldate   = {2021-09-30},
#   doi       = {10.1080/00031305.2016.1277159}
# }

#%%
import numpy as np
from scipy.stats import multivariate_normal

cov = np.array([[1, 0.2, 0.3],
                [0.2,0.8,-0.1],
                [0.3,-0.1,1.2]])
mean = np.array([0.1,-0.2, 0.3])

n = 10000
X = multivariate_normal.rvs(mean=mean, cov=cov, size=n)
mu = X.mean(axis=0)
X_bar = X - mu
C = np.cov(X_bar.T)
print(C)

L, V = np.linalg.eig(C) 
L = np.diag(L)
print(L)
print(V)

L_inv_squared = np.diag(np.diag(L)**(-0.5))
# print(L_inv_squared)
W = np.dot(np.dot(V, L_inv_squared),V.T)
print(W)
# print(np.dot(np.dot(V, L), V.T))

X_star = X_bar @ W

print(np.cov(X_star.T))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 26 11:28:25 2022

@author: frem
"""

# -*- coding: utf-8 -*-

from scipy import stats
import numpy as np
from matplotlib import pyplot as plt
import ekde
from time import time

import fastkde
import KDEpy

#%%
n = 10**6
d = 5

np.random.seed(42)
X = stats.multivariate_normal.rvs(mean=np.zeros(d), cov=np.diag(np.ones(d)), size=n)

np.random.seed(51)
Y = stats.multivariate_normal.rvs(mean=np.zeros(d), cov=np.diag(np.ones(d)), size=n)

np.random.seed(None)

kde_ekde = ekde.KDE(q=31, verbose=1).fit(X)

h = ekde.base.scotts_rule(X)

#%%
from sklearn.neighbors import KNeighborsRegressor

class KDEpy_FFTKDE():
    def __init__(self, 
                 h,
                 s,
                 q,
                 kernel='gaussian'):
        self.h = h
        self.s = s
        self.q = q
        self.kernel = kernel
        
        self.kde = KDEpy.FFTKDE(kernel=self.kernel, bw=self.h)
        
    def fit(self, X):
        self.kde.fit(X)
        
        self.n_grid = int(np.max(X.max(axis=0) - X.min(axis=0)) / (self.h * self.s / self.q))
        
    def predict(self, X):
        
        X_grid, f_grid = self.kde.evaluate(self.n_grid)
        
        knr = KNeighborsRegressor(n_neighbors=1).fit(X_grid, f_grid)
        
        return(knr.predict(X))
    
from fastkde import fastKDE
class FastKDE_process():
    def __init__(self,
                 h,
                 s,
                 q):
        self.h = h
        self.s = s
        self.q = q
        
    def fit(self, X):
        self.data = X
        
        self.n_grid = int(np.max(X.max(axis=0) - X.min(axis=0)) / (self.h * self.s / self.q))
        
        self.n_grid = 2**(int(np.log2(self.n_grid)) + 1) + 1
    
    def predict(self, X):
        
        axes = [np.linspace(x.min(), x.max(), self.n_grid) for x in X.T]
        
        fkde = fastKDE.fastKDE(data=self.data.T,
                               axes=axes)
        
        axes = np.meshgrid(*axes)
        X_grid = np.vstack([a.flat for a in axes]).T
        
        knr = KNeighborsRegressor(n_neighbors=1).fit(X_grid, fkde.pdf.flat)
        
        return(knr.predict(X))

#%%
st = time()
fftkde = KDEpy_FFTKDE(h=h,
                      s=3,
                      q=31)
fftkde.fit(X)
f_kdepy_fft = fftkde.predict(Y)
time_kdepy = time()-st
print(time_kdepy)
#%%
fkde = FastKDE_process(h=h, s=3, q=31)
fkde.fit(X)
f_fastkde = fkde.predict(Y)

#%%
kde_ekde = ekde.KDE(q=31, verbose=1).fit(X)
f_ekde = kde_ekde.predict(Y)

#%%
print(np.max(f_ekde - stats.multivariate_normal.pdf(Y, mean=np.zeros(d), cov=np.diag(np.ones(d)))))


# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt
import numpy as np

X = np.array([
              # [1.2,0.2],
              [1.3,1.2],
              [1.7,1.7],
              [2.4,1.7],
              [2.7,1.3],
              [2.5,2.2],
              [2.1,2.9],
               [2.8,1.6],
              # [3.2,1.3],
              [3.2,1.2],
              [3.8,2.4],
               [3.7, 2.3]
              # [3.7,2.3],
              # [4.6,2.5],
              # [4.2,3.2],
              ])

X += np.array([0, 1.0])

Y = np.array([[1.6, 2.2],
              # [2.1,1.2],
              # [3.8, 2.7],
              ])

fig, axs = plt.subplots(1,1)
axs.set_aspect('equal', 'box')

X_real = X.copy() + np.array([52,3])
Y_real = Y + np.array([52,3])

plt.scatter(X_real[:,0], X_real[:,1], label='observed points')
plt.scatter(Y_real[:,0], Y_real[:,1], label='point to estimate')
# plt.hlines(3.0, 53, 56, color='black', linewidth=3, label='border')
plt.legend(loc='lower left', bbox_to_anchor=(1,0.762))
plt.xlabel('$z_1$')
plt.ylabel('$z_2$')
plt.savefig('/home/frem/Work/LUCC/gisworknotes/1_Articles/2_Calibration/figures/ekde_1.pdf', bbox_inches='tight')

#%%
fig, axs = plt.subplots(1,1)
axs.set_aspect('equal', 'box')
plt.scatter(X[:,0], X[:,1], s=20, label='observed')

plt.axis('off')

for x in [0,1,2,3,4, 5]:
    if x == 0:
        plt.vlines(x, 0, 5, color='gray', label='bin cells')
    else:
        plt.vlines(x, 0, 5, color='gray')
    if x < 5:
        plt.text(x+0.5, -0.5, x)

for y in [0,1,2,3,4, 5]:
    plt.hlines(y, 0, 5, color='gray')
    if y < 5:
        plt.text(-0.5, y+0.5, y)

# plt.hlines(0, 0, 5, color='black', linewidth=3, label='border')

plt.legend(loc='lower left', bbox_to_anchor=(1,0.715))

#%%
X_star = np.array([
                    [1,1],
                    # [2,0],
                    [2,1],
                   [2,2],
                   [3,1],
                   [3,2],
                   # [4,2],
                   # [4,3],
                   ])
X_star += np.array([0,1])

mu = np.array([2, 3,2,1,2])
# mu = mu / X.shape[0]

fig, axs = plt.subplots(1,1)
axs.set_aspect('equal', 'box')

plt.axis('off')



for x in [0,1,2,3,4, 5]:
    if x == 0:
        plt.vlines(x, 0, 5, color='gray', label='bin cells\nwith its $\\nu$ value noted\n($0$ if any)')
    else:
        plt.vlines(x, 0, 5, color='gray')
    if x < 5:
        plt.text(x+0.5, -0.5, x)

for y in [0,1,2,3,4, 5]:
    plt.hlines(y, 0, 5, color='gray')
    if y < 5:
        plt.text(-0.5, y+0.5, y)

# plt.hlines(0, 0, 5, color='black', linewidth=3, label='border')

from matplotlib import patches
from matplotlib.collections import PatchCollection

R = []
for x in X_star:
    R.append(patches.Rectangle(x, 1, 1))

colors = mu / mu.sum()
collection = PatchCollection(R, cmap=plt.cm.autumn, alpha=0.5)
collection.set_array(-colors)
# axs.add_collection(collection)

for i, x in enumerate(X_star):
    plt.text(x[0]+0.4, x[1]+0.37, round(mu[i],1))

plt.scatter(X[:,0], X[:,1], s=20, label='observed points', alpha=1)
# plt.text(5.5, 1, 'Bin values\nrefers to $\mathbf{\mu}$')
plt.legend(loc='lower left', bbox_to_anchor=(1,0.6))
plt.savefig('/home/frem/Work/LUCC/gisworknotes/1_Articles/2_Calibration/figures/ekde_2.pdf', bbox_inches='tight')


#%%
# fig, axs = plt.subplots(1,1)
# axs.set_aspect('equal', 'box')

# plt.axis('off')

# for x in [0,1,2,3,4, 5]:
#     if x == 0:
#         plt.vlines(x, 0, 5, color='gray', label='bin cells')
#     else:
#         plt.vlines(x, 0, 5, color='gray')
#     if x < 5:
#         plt.text(x+0.5, -0.5, x)

# for y in [0,1,2,3,4, 5]:
#     plt.hlines(y, 0, 5, color='gray')
#     if y < 5:
#         plt.text(-0.5, y+0.5, y)

# plt.hlines(0, 0, 5, color='black', linewidth=3, label='border')

# plt.scatter(Y[0,0], Y[0,1], s=20, marker='x', color='black', label='to estimate')

# colors = {0:'tab:blue', 1:'tab:orange', 2:'tab:green'}
# for i in [0,1,2]:
    
#     Yx = Y[i,0]
#     Yy = Y[i,1]
    
#     plt.scatter(Yx, Yy, s=20, color=colors[i], marker='x')
    
# plt.scatter(X_star[:4,0]+0.5, X_star[:4,1]+0.5, marker='s', color='tab:blue')
# plt.scatter(X_star[4:,0]+0.5, X_star[4:,1]+0.5, marker='s', alpha=0.3, color='tab:blue')
# plt.scatter(0.1,6.45,marker='x', color='tab:blue')
# plt.vlines(0, 0,5,colors='tab:blue', ls='--', lw=3)
# plt.vlines(3, 0,5,colors='tab:blue', ls='--', lw=3)
# plt.text(0,5.5,'    exploring :\nfirst select all close enough centers\naccording to $x_0$')

# #%%
# fig, axs = plt.subplots(1,1)
# axs.set_aspect('equal', 'box')

# plt.axis('off')

# for x in [0,1,2,3,4, 5]:
#     if x == 0:
#         plt.vlines(x, 0, 5, color='gray', label='bin cells')
#     else:
#         plt.vlines(x, 0, 5, color='gray')
#     if x < 5:
#         plt.text(x+0.5, -0.5, x)

# for y in [0,1,2,3,4, 5]:
#     plt.hlines(y, 0, 5, color='gray')
#     if y < 5:
#         plt.text(-0.5, y+0.5, y)

# plt.hlines(0, 0, 5, color='black', linewidth=3, label='border')

# plt.scatter(Y[0,0], Y[0,1], s=20, marker='x', color='black', label='to estimate')

# colors = {0:'tab:blue', 1:'tab:orange', 2:'tab:green'}
# for i in [0,1,2]:
    
#     Yx = Y[i,0]
#     Yy = Y[i,1]
    
#     plt.scatter(Yx, Yy, s=20, color=colors[i], marker='x')
    
# plt.scatter(X_star[:3,0]+0.5, X_star[:3,1]+0.5, marker='s', color='tab:blue')
# plt.scatter(X_star[3:,0]+0.5, X_star[3:,1]+0.5, marker='s', alpha=0.3, color='tab:blue')
# plt.vlines(0, 0,5,colors='tab:blue', ls='--', lw=3)
# plt.vlines(3, 0,5,colors='tab:blue', ls='--', lw=3)
# plt.hlines(0, 0,5,colors='tab:blue', ls='--', lw=3)
# plt.hlines(2, 0,5,colors='tab:blue', ls='--', lw=3)
# plt.scatter(0.1,6.45,marker='x', color='tab:blue')
# plt.text(0,5.5,'    exploring :\nthen keep only close enough centers\naccording to $x_1$')

#%%
fig, axs = plt.subplots(1,1)
axs.set_aspect('equal', 'box')

plt.axis('off')

for x in [0,1,2,3,4, 5]:
    if x == 0:
        plt.vlines(x, 0, 5, color='gray', label='bin cells')
    else:
        plt.vlines(x, 0, 5, color='gray')
    if x < 5:
        plt.text(x+0.5, -0.5, x)

for y in [0,1,2,3,4, 5]:
    plt.hlines(y, 0, 5, color='gray')
    if y < 5:
        plt.text(-0.5, y+0.5, y)

# plt.hlines(0, 0, 5, color='black', linewidth=3, label='border')

plt.scatter(Y[0,0], Y[0,1], s=20, color='tab:orange', label='to estimate')

colors = {0:'tab:blue', 1:'tab:orange', 2:'tab:green'}
# for i in [0,1,2]:
    
#     Yx = Y[i,0]
#     Yy = Y[i,1]
    
#     plt.scatter(Yx, Yy, s=20, color=colors[i], marker='x')

for i, x in enumerate(X_star):
    plt.text(x[0]+0.4, x[1]+0.37, round(mu[i],2))

C = [1.5, 2.5]

plt.plot([C[0]-1.5, C[0] + 1.5, C[0] + 1.5, C[0] - 1.5, C[0] - 1.5],
          [C[1] - 1.5, C[1] - 1.5, C[1] + 1.5, C[1] + 1.5, C[1] - 1.5], color='tab:orange', ls='--', linewidth=3, label='kernel support')


plt.text(5.5, 3, '$\\tilde{p} = $', fontsize=20)
plt.text(6.5, 2.2, '$\\frac{2}{10} K(\\frac{\\sqrt{2}}{2} \\frac{2hs}{q})$', fontsize=20)
plt.text(6.1, 1.2, '$+ \\frac{3}{10} K(\\frac{2hs}{q})$', fontsize=20)
plt.text(6.1, 0.2, '$+ \\frac{2}{10} K(0)$', fontsize=20)

plt.arrow(2.7,3.3,6.5-2.7-0.1,2.2-3.3+0.1, shape='full', width=0.07, facecolor='gray', edgecolor='none', length_includes_head=True)

plt.arrow(2.7,2.3,6.1-2.7-0.1,1.2-2.3+0.1, shape='full', width=0.07, facecolor='gray', edgecolor='none', length_includes_head=True)

plt.arrow(1.7, 2.3, 6.1-1.7-0.1,0.2-2.3+0.3, shape='full', width=0.07, facecolor='gray', edgecolor='none', length_includes_head=True)

plt.legend(loc='lower left', bbox_to_anchor=(1,0.71))
plt.savefig('/home/frem/Work/LUCC/gisworknotes/Articles/LUCCProbabilitiesEstimation_EnvModSoft/figures/ekde_3.pdf', bbox_inches='tight')


# -*- coding: utf-8 -*-

from scipy import stats
import numpy as np
from matplotlib import pyplot as plt
import ekde
from time import time

#%% Dataset

X_min = np.array([-0.5, 0.3])
x1_max = 1.5

mean = np.array([0,0])
rho = 0.7
cov = np.array([[1,rho],[rho,1]])

def _pdf(X, X_min):
    f = stats.multivariate_normal.pdf(X, mean=mean, cov=cov) 

    f[np.any(X<X_min, axis=1)] = 0
    f[X[:,1]>=x1_max] = 0

    return(f)

def _rvs(X_min, n,seed=None):
    np.random.seed(seed)
    X = stats.multivariate_normal.rvs(mean=mean, cov=cov, size=n)
    np.random.seed(None)
    X = X[np.all(X >= X_min, axis=1)]
    X = X[X[:,1] < x1_max]

    return(X)

def bounded_set(n, seed):
    X = _rvs(X_min, n, seed=seed)
    
    xk = (np.linspace(X[:,0].min()-X[:,0].std(),X[:,0].max()+X[:,0].std(),400),
          np.linspace(X[:,1].min()-X[:,1].std(),X[:,1].max()+X[:,1].std(),400))
    X_grid = np.meshgrid(xk[0], xk[1])
    X_grid = np.vstack((X_grid[0].flat, X_grid[1].flat)).T
    
    pdf_grid = np.sum(_pdf(X_grid, X_min)) * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0]
    
    Y = np.vstack((np.ones(100)*0.5,
                   np.linspace(0,2,100))).T
    
    pdf_Y = _pdf(Y, X_min) / pdf_grid
    
    return(X, Y, pdf_Y, X_grid)
    
def hyp_func(hyp, x):
    # a0 x0 + a1 x1 + r = 0
    # x1 = -r - a0 x0 / a1
    return((-hyp.r - hyp.a[0] * x) / hyp.a[1])

# %%
X, Y, pdf_Y, X_grid = bounded_set(10**7, 42)
print(X.shape[0]/10**6)
pdf_grid = np.sum(_pdf(X_grid, X_min)) * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0]
f_grid_exact = _pdf(X_grid, X_min) / pdf_grid

#%%
st = time()
bkde = ekde.KDE(h='terrel', 
                q=11,
                 kernel='box',
                    bounds=['left', 'both'],
                    n_jobs=1,
                    n_mc_axes = 10,
                    wt = True,
                    verbose=0)
# bkde.fit(np.array([[1,1]]))
bkde.fit(X)
print(time()-st)
# print(bkde._normalization)

from time import time
st = time()
f_grid = bkde.predict(X_grid)
print('predict time', time()-st)

print('I', f_grid.sum() * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0])

fig, axs = plt.subplots(1,1)
axs.set_aspect('equal')
plt.scatter(X_grid[:,0], X_grid[:,1], c=f_grid, s=2)
plt.colorbar()

print('mad', np.mean(np.abs(f_grid - f_grid_exact)))

#%%
n = X.shape[0]
Z = X - np.mean(X, axis=0)

U, S, Vt = np.linalg.svd(Z, full_matrices=False)

C = np.cov(Z.T)

L, V2 = np.linalg.eig(C)

#%%
Sigma = np.diag(S)
Sigma_inv = np.linalg.inv(Sigma)
V = Vt.T

print(np.linalg.det(Sigma_inv @ V))
print(1/ np.linalg.det(Sigma @ Vt))

#%%
print(1/ bkde._wt.scale_)
print(1 / np.linalg.det((n-1)**(-1/2) * Sigma @ Vt))
print(1 / (n-1)**(-d/2) / np.linalg.det(Sigma @ Vt))
print((n-1)**(d/2) * np.linalg.det(Sigma_inv @ V))

#%%
import KDEpy

# st = time()
# y = KDEpy.TreeKDE(kernel='box', bw=bkde._h).fit(X).evaluate(int(np.sqrt(X_grid.shape[0])))
# print('Tree time', time()-st)
# print('mad', np.mean(np.abs(y[1] - _pdf(y[0], X_min) / pdf_grid)))

st = time()
y = KDEpy.FFTKDE(kernel='box', bw=bkde._h).fit(X).evaluate(int(np.sqrt(X_grid.shape[0])))
print('FFT time', time()-st)
print('mad', np.mean(np.abs(y[1] - _pdf(y[0], X_min) / pdf_grid)))
#%%
mad = []
q_list = np.arange(1, 51,2)
for q in q_list:
    print(q)
    bkde = ekde.KDE(h='terrel', 
                    q=q,
                     kernel='box',
                        bounds=[
                            (0, 'left'),
                            (1, 'both'),
                            ],
                        n_jobs=2,
                        n_mc_axes = 10,
                        wt = True,
                        verbose=0)
    
    bkde.fit(X)
    f_grid = bkde.predict(X_grid)
    mad.append(np.mean(np.abs(f_grid - f_grid_exact)))

#%%
plt.plot(q_list, mad)
plt.xlabel('q')
plt.ylabel('MAD')

#%%
d = 2
n = 10**4
q = 101
n_axes = 300
kernel='box'
wt = True
h_select = 'terrel'

np.random.seed(42)
X = stats.multivariate_normal.rvs(mean=np.zeros(d), cov=np.diag(np.ones(d)), size=n)

if d == 1:
    X = X[:,None]

np.random.seed(35)
Y = stats.multivariate_normal.rvs(mean=np.zeros(d), cov=np.diag(np.ones(d)), size=n)
if d == 1:
    Y = Y[:,None]
    
f_Y = stats.multivariate_normal.pdf(Y, mean=np.zeros(d), cov=np.diag(np.ones(d)))

np.random.seed(None)

kde = ekde.KDE(h = h_select,
                q=q,
                kernel=kernel,
                n_jobs=2,
                n_mc_axes = 10,
                wt=wt,
                verbose=0).fit(X)

h_t = kde._h
print(h_t)

mse = []
list_h = np.arange(h_t*0.5, h_t * 2, h_t * 0.1)
for i, h in enumerate(list_h):
    print(round(i/list_h.shape[0],2), h)
    kde = ekde.KDE(h = h,
                   q=q,
                   kernel=kernel,
                   wt=wt,
                   verbose=0)
    kde.fit(X)
    f = kde.predict(Y)
    
    mse.append(np.abs(f - f_Y).mean())

plt.plot(list_h, mse)
plt.vlines(h_t, np.min(mse), np.max(mse))

# pandas -> numpy vérifier temps de calcul
# multijobs ?

#%%
X_grid = ekde.base.generate_grid(*[np.linspace(-4, 4, n_axes) for j in range(d)])
print((stats.multivariate_normal.pdf(X_grid, mean=np.zeros(d), cov=np.diag(np.ones(d)))).sum() * 8**d / X_grid.shape[0])
print((stats.multivariate_normal.pdf(X_grid, mean=np.zeros(d), cov=np.diag(np.ones(d)))**2).sum() * 8**d / X_grid.shape[0])
print(1 / (2 * np.sqrt(np.pi))**d)

#%%
mu = 1 / (2 * np.sqrt(np.pi))**d
h_terrel = (R / (X.shape[0] * mu))**(1/5)
print(h_terrel)

#%% box convolution computation


#%%
plt.plot(np.arange(0.04, 0.3, 0.01), mse)

#%%

margin = (bkde.q-1)/2

i_Z = 0

AB = []
for j in range(2):
    sort_idx = np.argsort(bkde._U[:,j])
    u = bkde._U[sort_idx, j].astype(np.intc)
    a = ekde.ekdefunc.np_binary_search_left(u, Z[i_Z,j] - margin, 0, u.shape[0])
    b = ekde.ekdefunc.np_binary_search_right(u, Z[i_Z,j] + margin, a, u.shape[0])
    
    AB.append(sort_idx[a:b+1])
    
AB = np.hstack(AB)
AB = np.sort(AB)

s = 0
d = 2
out = 0
for i in range(1, AB.size):
    if AB[i] == AB[i-1]:
        out += 1
    else:
        out = 0
    if out == d-1:
        s += bkde._nu[AB[i]]
        
#%%
z = Z[:,0].astype(np.intc)

ekde.ekdefunc.quickSort(arr, 0, z.size-1)

#%%

st = time()
Z_s_np = np.sort(z)
print(time()-st)

st = time()
Z_s_ekde = ekde.ekdefunc.quickSort(z, 0, z.size-1)
print(time()-st)
        
#%%
Z = np.array([[1,1],
              [1,3],
              [1,4],
              [2,4],
              [3,3],
              [1,1],
              [1,2],
              [2,2],
              [4,4]])

AB = []
for j in range(2):
    sort_idx = np.argsort(Z[:,j])
    z = Z[sort_idx, j].astype(np.intc)
    a = ekde.ekdefunc.np_binary_search_left(z, 2, 0, Z_sbc_0.shape[0])
    b = ekde.ekdefunc.np_binary_search_right(z, 3, a, Z_sbc_0.shape[0])

#%%
x = np.array([12,14,11,11,13,15,14]).astype(np.intc)
sort_idx = np.argsort(x)
sort_inverse_idx = np.argsort(sort_idx)

y = x[sort_idx]

z = y[sort_inverse_idx]

a = ekde.ekdefunc.np_binary_search_left(y, 12, 0, 7)
b = ekde.ekdefunc.np_binary_search_right(y, 14, 0, 7)

# s = np.zeros(b-a+1)
# for i in range(b-a+1):
#     s[sort_idx[i]] = x[sort_idx[a + i]]

print(x)
print(sort_idx)
print(y)
print(z)
print(a, b)
print(np.arange(7)[a:b+1])

#%%
Z_sbc_1 = Z[sort_idx[1], 1]
plt.plot(Z_sbc_1)

#%%
plt.plot(Z[Z_sbc_idx[1], 1])

#%%
i_Z = 0
test_inf = np.all(bkde._U >= Z[i_Z] - margin, axis=1)
test_sup = np.all(Z[i_Z] + margin >= bkde._U, axis=1)
select = np.all(np.vstack((test_inf, test_sup)), axis=0)

f_0 = np.sum(bkde._nu[select]) 
# f_0 /= (bkde._normalization * bkde._wt.scale_)
print(f_0)

#%%
idx_0 = np.where(np.all((Z[i_Z,0] - margin <= bkde._U[:,0], 
                         Z[i_Z,0] + margin >= bkde._U[:,0]), axis=0))[0]
idx_01 = np.where(np.all((Z[i_Z,1] - margin <= bkde._U[idx_0,1], 
                          Z[i_Z,1] + margin >= bkde._U[idx_0,1]), axis=0))[0]

f_0 = np.sum(bkde._nu[idx_0[idx_01]])
# f_0 /= (bkde._normalization * bkde._wt.scale_)
print(f_0)

#%%
f = bkde.predict(X)
print(f[0])

#%%
X2, Y, pdf_Y, X_grid = bounded_set(10**6, 30)
idx = np.random.choice(X2.shape[0], 10**4, replace=False)
#%%
# bkde.fit(X[:10])
from time import time
st = time()
f_X = bkde.predict(X2)
print(time()-st)
plt.scatter(X2[idx,0], X2[idx,1], c=f_X[idx], s=2)
plt.scatter(X[:10,0], X[:10,1], c='red')



#%%
# U = np.array([[1,1,1],
#               [1,1,3],
#               [1,1,4],
#               [1,2,4],
#               [1,3,3],
#               [2,1,1],
#               [2,1,2],
#               [2,2,2]], dtype=np.intc)

# U_diff_desc = np.ones(U.shape, dtype=np.intc)
# ekde.ashfunc.count_diff_desc(U, U_diff_desc)

# bkde._U = U
# bkde._U_diff_desc = U_diff_desc

#%%
f_X = bkde.predict(X)
# 10**7 -> 42 sec 
# 10**6 -> 3.3 sec
# ([4, 249], 11.652764262505263)

#%%
f_eval = bkde.predict(Y)
plt.plot(Y[:,1], pdf_Y)
plt.plot(Y[:,1], f_eval)

#%%
f_X = bkde.predict(X)
print('mad', np.abs(f_X - _pdf(X, X_min) / pdf_grid).mean())

#%%
idx = np.random.choice(X.shape[0], 10**4, replace=False)
plt.scatter(X[idx,0], X[idx,1], c=bkde.predict(X[idx]), s=2)

#%%
import pandas as pd
Z = ekde.ash.discretize(bkde._wt.transform(X), bkde._x_min, bkde._h/bkde.q)
n_grid, d = Z.shape
Z = pd.DataFrame(Z)

#%%
Z['j'] = np.arange(n_grid)
Z = Z.sort_values(by=[i for i in range(X_grid.shape[1])])
Z_indices = Z['j'].values.astype(np.intc)
Z = Z[[i for i in range(d)]].values.astype(np.intc)
Z_diff = np.ones((n_grid, d), dtype=np.intc)
ekde.ashfunc.count_diff(Z, Z_diff)
f = np.array(ekde.ashfunc.merge(U=bkde._U,
                                   nu=bkde._nu,
                                   Z=Z,
                                   Z_indices=Z_indices,
                                   Z_diff=Z_diff,
                                   q=bkde.q))
#%%
print(f_grid.sum() * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0])
pdf_grid = np.sum(_pdf(X_grid, X_min)) * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0]
f_grid_exact = _pdf(X_grid, X_min) / pdf_grid
print(f_grid_exact.sum() * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0])

#%%
mad = []
for h in np.arange(0.1, 0.5, 0.02):
    
    bkde = ekde.BKDE(
                        h = float(h),
                        q=101,
                        bounds=[
                                (0, 'left'),
                                (1, 'both'),
                                ],
                        )
    bkde.fit(X)
    f_grid = bkde.predict(X_grid)
    
    mad.append(np.abs(f_grid - f_grid_exact).mean())

#%%
plt.plot(np.arange(0.1,0.5,0.02), mad)



#%%
a = ekde.ash.discretize(bkde._wt.transform(Y[[17],:]), bkde._x_min, dx=bkde._h/bkde.q)

#%%
plt.scatter(bkde._U[:,0], bkde._U[:,1], s=0.1, c=bkde._nu)
# plt.xlim([0,1000])
# plt.ylim([0,500])
plt.colorbar()

#%%
from hyperclip import Hyperplane
from hyperclip import hyperfunc, Hyperclip
h = 0.3165675288885327
z = np.array([[14,40]])
x = ekde.ash.compute_centers(z, bkde._x_min, bkde._h/bkde.q)[0]
a = x - bkde._h/2
b = x + bkde._h/2
x_r = np.random.random((1000,2)) * (b-a) + a
A = np.zeros((2,2))
R = np.zeros(2)
is_in = np.ones(x_r.shape[0]).astype(bool)
for i, i_hyp in enumerate([1]):
    hyp = Hyperplane(bkde.A[:,i_hyp], bkde.R[i_hyp])
    
    is_in = np.all((is_in, hyp.side(x_r)), axis=0)
    
    hyp.affine_transform(translation = - x + bkde._h/2, 
                         scale = np.ones(bkde._d) / bkde._h,
                         inplace=True)
    # hyp.set_positive_side(x)
    A[:, i] = hyp.a
    R[i] = hyp.r

# print
hyperclip = Hyperclip(cython=True).set_A_R(A, R)
# print('class', hyperclip.volume())

A = np.hstack((A, np.zeros((2,1))))
R = np.hstack((R, 0))
print("check A", hyperfunc.clipping_condition_A_according_m(hyperclip.A, hyperclip.R, m=2))
print('func', hyperfunc.volume_according_m(hyperclip.A, hyperclip.R, m=2, check_A=False, zero=hyperclip.zero))
# print(ekde.ashfunc.volume(hyperclip.A, hyperclip.R, check_A=False))
print(is_in.mean())
plt.scatter(x_r[:,0], x_r[:,1], c=is_in)
plt.colorbar()

#%%
from hyperclip import Hyperplane
i_hyp = 0

x = np.array([-1.57648396, -2.72955008])
hyp = Hyperplane(bkde.A[:,i_hyp], bkde.R[i_hyp])
a = x - bkde._h/2
b = x + bkde._h/2
x_r = np.random.random((1000,2)) * (b-a) + a
plt.scatter(x[0], x[1])
plt.plot([x[0] - bkde._h/2, x[0] + bkde._h/2],
         [hyp_func(hyp, x[0] - bkde._h/2), hyp_func(hyp, x[0] + bkde._h/2)])
plt.xlim([x[0] - bkde._h/2, x[0] + bkde._h/2])
plt.ylim([x[1] - bkde._h/2, x[1] + bkde._h/2])
plt.scatter(x_r[:,0], x_r[:,1], c=hyp.side(x_r).astype(int))
plt.colorbar()
plt.show()

#%%
hyp.affine_transform(translation = - x + bkde._h/2, 
                     scale = np.ones(bkde._d) / bkde._h,
                     inplace=True)
x = np.array([0.5,  0.5])
plt.scatter(x[0], x[1])
plt.plot([0, 1],
         [hyp_func(hyp, 0), hyp_func(hyp, 1)])
plt.xlim([0,1])
plt.ylim([0,1])
print(hyp.a, hyp.r)



#%%
# h = bkde._h
wt = bkde._wt
plt.scatter(wt.transform(X)[:,0], wt.transform(X)[:,1])



#%%
f_eval = bkde.predict(Y)
plt.plot(Y[:,1], pdf_Y)
plt.plot(Y[:,1], f_eval)

#%%
f_grid = bkde.predict(X_grid)
print(f_grid.sum() * np.product(X_grid.max(axis=0) - X_grid.min(axis=0)) / X_grid.shape[0])
print(_pdf(X_grid, X_min).sum() * np.product(X_grid.max(axis=0) - X_grid.min(axis=0)) / X_grid.shape[0])

#%%
# dx = bkde._h / bkde.q
# Y = bkde._x_min + bkde._U * dx + 0.5 * dx
# # Y = bkde._wt.inverse_transform(Y)
# plt.scatter(Y[:,0], Y[:,1], c=bkde._nu, s=2)
# plt.colorbar()

#%%
st = time()
f = bkde.predict(X)
print(time()-st)



#%%
ash = ekde.ASH(q=10,
        bounds=[
                # (0, 'left'),
                # (1, 'both')
                ],
        n_mc=10000,
        n_jobs=4)
st = time()

# ash._compute_bandwidth(X[:10**6])
# h = ash._h

#%%
st = time()
ash.fit(X)
print('fit exec time : ', time()-st)

#%%
plt.scatter(ash._U_nu[0], ash._U_nu[1], c=ash._U_nu["nu"])
# plt.xlim([10,15])
# plt.ylim([2,5])

#%%
# X = np.array([[0.2,0.4]])
st = time()
f = ash.predict(X)
print(time()-st)
# print(f)
# plt.plot(Y[:,1], pdf_Y)
# plt.plot(Y[:,1], f)
# plt.show()
plt.scatter(X[:,0], X[:,1],  c=f[1], s=4)
#%%
a = np.meshgrid(*(np.arange(10) for i in range(3)))
a = np.vstack([aa.flat for aa in a]).T

#%%
plt.scatter(X[:,0], X[:,1],  c=f[1], s=4)

#%%
plt.scatter(f[0][:,0], f[0][:,1], c=np.arange(f[0].shape[0]))



#%%
X = np.array([[1.1, 1.1],
              [1.1, 1.2],
              [2.1, 1.1]])
ash = ekde.ASH(h=3.0, q=3)
ash.fit(X)

#%%
X_eval = np.array([[0.5, 1.1],
                   [0.5, 0.5]])
f = ash.predict(X_eval)
print(f)
#%%
U_N = ash._U_N[0]

U = U_N[[i for i in range(X.shape[1])]].values
N = U_N["N"].values

i_shift = 0
U_shift = ((U - i_shift) / 100).astype(int)
print(U_shift)

#%%
%timeit ash.fit(X)
# cython : 174 ms
# npg : 103 ms


#%%
# X = np.array([[]])

#%%
import pandas as pd

class Digitize():
    def __init__(self, dx, shift=0):
        self.dx = dx
        self.shift = shift

    def fit(self, X):
        self._d = X.shape[1]
        self._bins = [np.arange(V.min() - self.dx + self.shift,
                                V.max() + self.dx + self.shift,
                                self.dx) for V in X.T]

        return (self)

    def transform(self, X):
        X = X.copy()
        for k in range(self._d):
            X[:, k] = np.digitize(X[:, k], bins=self._bins[k])
        return (X.astype(int))

    def fit_transform(self, X):
        self.fit(X)

        return (self.transform(X))

def pandas_func(X):
    X_wt = ash._wt.fit_transform(X)
    
    histograms = []
    for i_shift in range(ash.q):
        digitizer = Digitize(dx = ash._h, shift = i_shift * ash._h / ash.q) 
        df = pd.DataFrame(digitizer.fit_transform(X_wt))
        
        df_uniques = df.groupby(by=df.columns.to_list()).size().reset_index(name='P')
        df_uniques['P'] /= df_uniques['P'].sum()
        
        histograms.append(df_uniques)
    return(histograms)

def numpy_func(X):
    X_wt = ash._wt.fit_transform(X)
    
    histograms = []
    for i_shift in range(ash.q):
        digitizer = Digitize(dx = ash._h, shift = i_shift * ash._h / ash.q) 
        X_wt_digit = digitizer.fit_transform(X_wt)
        

st = time()
# histograms = numpy_func(X)
X_wt = ash._wt.fit_transform(X)

histograms = []
for i_shift in range(ash.q):
    digitizer = Digitize(dx = ash._h, shift = i_shift * ash._h / ash.q) 
    X_wt_digit = digitizer.fit_transform(X_wt)
    histograms.append(np.unique(X_wt_digit, axis=0, return_counts=True))
    
print(time()-st)

#%%
import numpy_indexed as npi
import numpy_groupies as npg
st = time()
npga = npg.aggregate(X_wt_digit.T, a=1, func='sum', fill_value=0)
print(time()-st)

#%%
import numpy_groupies as npg
npga = npg.aggregate(X_wt_digit.T, a=1, func='sum', fill_value=0)

#%%
%timeit pandas_func(X)

#%%
from numpy_indexed import *
# three sets of graph edges (doublet of ints)
edges = np.random.randint(0, 9, (3, 100, 2))
# find graph edges exclusive to one of three sets
ex = exclusive(*edges)
print(ex)
# which edges are exclusive to the first set?
print(contains(edges[0], ex))
# where are the exclusive edges relative to the totality of them?
print(indices(union(*edges), ex))
# group and reduce values by identical keys
values = np.random.rand(100, 20)
# and so on...
print(group_by(edges[0]).median(values))

#%%

# st = time()
# n_X = np.max([ash._fit_results[i_shift][0].shape[0] for i_shift in range(ash.q)])

# ash._X_digit_uniques = np.zeros((ash.q, n_X, ash._d))
# ash._P = np.zeros((ash.q, n_X))
# for i_shift in range(ash.q):
#     n_X_i = ash._fit_results[i_shift][0].shape[0]
    
#     ash._X_digit_uniques[i_shift, :n_X_i, :] = ash._fit_results[i_shift][0]
#     ash._P[i_shift, :n_X_i] = ash._fit_results[i_shift][1]

# ash._X_digit_uniques = ash._X_digit_uniques.astype(np.int32)

# print('self process exec time :', time()-st)

# print(len(ash._ret))
# f = ash.predict(X_grid)

# I = np.sum(f) * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0]
# print(I)

# st = time()
# f_grid = ash.predict(X_grid)
# print('grid exec time :', time()-st)

#%%
X_eval = np.vstack((np.ones(100)*0.5,
                        np.linspace(0,2,100))).T

st = time()
f_eval = ash.predict(X_eval)
print('eval exec time :', time()-st)

plt.plot(Y[:,1], pdf_Y)
plt.plot(Y[:,1], f_eval)
plt.show()

#%%
from sklearn.neighbors import KernelDensity
kde = KernelDensity(kernel='gaussian', bandwidth=ash._h / 2.576).fit(X)

st = time()
f_eval_kde = np.exp(kde.score_samples(X_eval))
print('eval exec time :', time()-st)

plt.plot(Y[:,1], pdf_Y)
plt.plot(Y[:,1], f_eval)
plt.plot(Y[:,1], f_eval_kde)
plt.show()



import numpy as np

import ekde.ekdefunc
from ekde.ekdefunc import merge as ekdefunc_merge
import hyperclip.hyperfunc
from hyperclip import Hyperplane
from .whitening_transformer import WhiteningTransformer
import pandas as pd
from scipy.special import gamma, factorial2
from scipy.stats import norm
from joblib import dump, load
import time
import os
import shutil
from joblib import wrap_non_picklable_objects
from multiprocessing import Pool

kernels_id = {'box' : 0,
              'gaussian' : 1,
              'epa' : 2}

supports = {'box' : 1.0,
            'gaussian' : 3.0,
            'epa' : 1.0}

R = {'box' : 1/2,
     'gaussian' : 1 / (2 * np.sqrt(np.pi)),
     'epa' : 3 / 5}

def compute_centers(Z, x_min, dx):
    C = x_min + Z * dx + 0.5 * dx
    return(C)

class KDE():
    def __init__(self, 
                 h='terrel', 
                 kernel='box',
                 q=51, 
                 bounds=[],
                 n_jobs = 1,
                 zero = 0.00000000000001,
                 n_mc_axes = 100,
                 wt = True,
                 verbose=0):
        if q%2 == 0:
            raise(ValueError("Unexpected q value. q should be an odd int."))
                
        self.h = h
        self.kernel = kernel
        self._h = None
        self.q = q
        self.bounds = bounds
        self.n_jobs = n_jobs
        self.zero = zero
        self.n_mc_axes = n_mc_axes
        self.wt = wt
        self.verbose = verbose
    
    def fit(self, X):
        if self.verbose > 0:
            print('KDE fit process, X.shape=', X.shape)
        # preprocessing
        if len(X.shape) == 1:
            X = X[:,None]
        
        self._real_X_min = np.min(X, axis=0)
        self._real_X_max = np.max(X, axis=0)
        
        # get data dimensions
        self._n, self._d = X.shape
        
        # find a point not on boundaries
        id_x = self._find_point_not_on_boundaries(X)
        
        # preprocessing
        if self.wt:
            self._wt = WhiteningTransformer()
            X = self._wt.fit_transform(X)
        
        self._x_min = np.min(X, axis=0)
        
        # BOUNDARIES INFORMATIONS
        A, R = self._set_boundaries(x = X[id_x])
        self.A = A
        self.R = R
        # BANDWIDTH SELECTION
        self._compute_bandwidth(X)
        
        self._support = supports[self.kernel]
        
        self._x_min = X.min(axis=0)
        self._dx = self._h * 2 * self._support / self.q
        
        Z = self._discretize(X).astype(np.intc)
        Z = pd.DataFrame(Z)
        
        U_nu = Z.groupby(by=Z.columns.tolist()).size().reset_index(name="nu")
        
        self._U = U_nu[Z.columns.tolist()].values.astype(np.intc)
        self._nu = U_nu["nu"].values.astype(np.double)
        
        self._U_diff_desc = np.ones((self._U.shape[0], self._d), dtype=np.intc)
        ekde.ekdefunc.count_diff_desc(self._U, self._U_diff_desc)
        
        C = compute_centers(self._U, self._x_min, self._dx).astype(np.double)
        
        if self.kernel == 'box':
            self._nu /= np.array(hyperclip.hyperfunc.volumes(A, 
                                                             R, 
                                                             C, 
                                                             self._h, 
                                                             zero=self.zero))
        elif self.kernel == 'gaussian':
            self._nu /= np.array(ekde.ekdefunc.boundary_correction_gaussian_kernel(A = A, 
                R = R, 
                C = C, 
                h = self._h, 
                n_mc=self.n_mc_axes**self._d,
                zero=self.zero))
        
        self._compute_normalization()
        
        return(self)
    
    def predict(self, X):
        if self.wt:
            X = self._wt.transform(X)
        
        id_out_of_bounds = np.zeros(X.shape[0]).astype(np.bool)
        for hyp in self._bounds_hyperplanes:
            id_out_of_bounds = np.any((id_out_of_bounds, ~hyp.side(X)), axis=0)
        Z = self._discretize(X)
        
        # SORT Z
        # ======
        Z = pd.DataFrame(Z)
        # keep the index safe
        Z['j'] = np.arange(Z.shape[0])
        # sort by i 
        Z = Z.sort_values(by=[i for i in range(self._d)])
        Z_indices = Z['j'].values.astype(np.intc)
        Z = Z[[i for i in range(self._d)]].values.astype(np.intc)
        
        Z_diff_asc = np.ones((Z.shape[0], self._d), dtype=np.intc)
        ekde.ekdefunc.count_diff_asc(Z, Z_diff_asc)
        
        id_Z_unique = np.where(Z_diff_asc[:, self._d-1] == 1)[0]
        
        streams = [int(id_Z_unique.size / self.n_jobs) * i for i in range(self.n_jobs + 1)]
        streams.append(id_Z_unique.size)
        
        folder = './.temp_joblib_memmap'
        try:
            os.mkdir(folder)
        except FileExistsError:
            pass
        
        Z_filename_memmap = os.path.join(folder, 'Z_memmap')
        dump(Z, Z_filename_memmap)
        Z_memmap = load(Z_filename_memmap, mmap_mode='r')
        
        U_filename_memmap = os.path.join(folder, 'U_memmap')
        dump(self._U, U_filename_memmap)
        U_memmap = load(U_filename_memmap, mmap_mode='r')
        
        U_diff_desc_filename_memmap = os.path.join(folder, 'U_diff_desc_memmap')
        dump(self._U_diff_desc, U_diff_desc_filename_memmap)
        U_diff_desc_memmap = load(U_diff_desc_filename_memmap, mmap_mode='r')
        
        nu_filename_memmap = os.path.join(folder, 'nu_memmap')
        dump(self._nu, nu_filename_memmap)
        nu_memmap = load(nu_filename_memmap, mmap_mode='r')
        
        pool = Pool(processes=self.n_jobs)
        g = pool.starmap(ekde.ekdefunc.merge, 
                          [(U_memmap,
                            U_diff_desc_memmap,
                            nu_memmap,
                            Z_memmap[id_Z_unique[streams[i_stream]: streams[i_stream+1]]],
                            self.q,
                            self._h,
                            kernels_id[self.kernel],
                            self._dx,
                            self.verbose) for i_stream in range(self.n_jobs + 1)])
        
        pool.close()
        
        try:
            shutil.rmtree(folder)
        except:  # noqa
            print('Could not clean-up automatically.')
        
        g = np.hstack([gi for gi in g])
        
        
        f = np.array(ekde.ekdefunc.set_estimation(Z_diff_asc=Z_diff_asc,
                                                  Z_indices=Z_indices,
                                                  g=g))
        
        f[id_out_of_bounds] = 0.0
        
        f = f / self._normalization
        
        if self.wt:
            f /= self._wt.scale_
            
        return(f)
    
    def _discretize(self, X):
        Z = ((X - self._x_min) / self._dx).astype(int)
        return(Z)
    
    def _find_point_not_on_boundaries(self, X):
        trigger_point_found = True
        
        X_min = np.min(X, axis=0)
        X_max = np.max(X, axis=0)
        
        while trigger_point_found:
            id_x = np.random.choice(self._n)
            x = X[id_x]
            
            if np.all(X_min < x) and np.all(x < X_max):
                trigger_point_found = False
                return(id_x)
        
        
    
    def _set_boundaries(self, x):
        self._bounds_hyperplanes = []
        
        if type(self.bounds) is list:
            for k, pos in enumerate(self.bounds):
                if pos == 'left':
                    self._add_boundary(k=k,
                                       value=self._real_X_min[k],
                                       x=x)
                elif pos == 'right':
                    self._add_boundary(k=k,
                                       value=self._real_X_max[k],
                                       x=x)
                elif pos == 'both':
                    self._add_boundary(k=k,
                                       value=self._real_X_min[k],
                                       x=x)
                    self._add_boundary(k=k,
                                       value=self._real_X_max[k],
                                       x=x)
                # else:
                    # raise(TypeError('Unexpected bounds parameters'))
        
        A = np.zeros((self._d, len(self._bounds_hyperplanes)))
        R = np.zeros(len(self._bounds_hyperplanes))
        
        for i_hyp, hyp in enumerate(self._bounds_hyperplanes):
            A[:, i_hyp] = hyp.a
            R[i_hyp] = hyp.r
        
        return(A, R)
    
    def _add_boundary(self, k, value, x):
        P = np.diag(np.ones(self._d))
        
        P[:, k] = value
        
        if self.wt:
            P = self._wt.transform(P)

        hyp = Hyperplane().set_by_points(P)
        hyp.set_positive_side(x)
        self._bounds_hyperplanes.append(hyp)
    
    def _compute_bandwidth(self, X):
        if type(self.h) is int or type(self.h) is float or type(self.h) is np.float64:
            self._h = float(self.h)

        elif type(self.h) is str:
            if self.h == 'scott' or self.h == 'silverman':
                # the scott rule is based on gaussian kernel
                
                self._h = scott_bandwidth(n=X.shape[0],
                                          d=X.shape[1])
                
                if self.kernel == 'box':
                    # the support of the gaussian kernel to have 99%
                    # of the density is (2.576/2) = 1.288
                    self._h *= 1.288
                elif self.kernel == 'gaussian':
                    self._h *= 4
            
            elif self.h == 'terrel':
                self._h = terrel_bandwidth(n=X.shape[0], 
                                           d=X.shape[1], 
                                           R=R[self.kernel]**self._d)
            
            else:
                raise (ValueError("Unexpected bandwidth selection method."))
        else:
            raise (TypeError("Unexpected bandwidth type."))

        if self.verbose > 0:
            print('Bandwidth selection done : h=' + str(self._h))
    
    
    
    def _compute_normalization(self):
        if self.kernel == 'box':
            self._normalization = (2 * self._h)** self._d
            
        elif self.kernel == 'gaussian':
            self._normalization = (self._h * np.sqrt(2 * np.pi))** self._d
            
        elif self.kernel == 'epa':
            self._normalization = self._h** self._d * volume_unit_ball(self._d)
        
        self._normalization = self._n * self._normalization 
        
            
    def set_params(self, **params):
        """
        Set parameters.

        Parameters
        ----------
        **params : kwargs
            Parameters et values to set.

        Returns
        -------
        self : DensityEstimator
            The self object.

        """
        for param, value in params.items():
            setattr(self, param, value)

def volume_unit_ball(d, p=2):
    # from KDEpy
    return 2.0 ** d * gamma(1 + 1 / p) ** d / gamma(1 + d / p)

def scott_bandwidth(n, d):
    """
    Scott's rule according to "Multivariate density estimation", Scott 2015, p.164.
    The Silverman's rule is exactly the same in "Density estimation for statistics and data analysis", Silverman 1986, p.87.
    Parameters
    ----------
    X : numpy array of shape (n_samples, n_features).

    Returns
    -------
    h : float
        Scotts rule bandwidth. The returned bandwidth should be then factored
        by the data variance.

    """
    return(n**(-1/(d + 4)))

def terrel_bandwidth(n, d, R):
    num = (d + 8) ** ((d+6)/2) * np.pi ** (d / 2) * R
    den = 16 * n * gamma((d+8)/2) * (d + 2)
    return (num / den)**(1/(d+4))

def generate_grid(*kwargs):
    xx = np.meshgrid(*kwargs)
    
    xx = np.vstack([xi.flat for xi in xx]).T
    
    return(xx)


# -*- coding: utf-8 -*-

import numpy as np
import ekde

#%%
folder = "/home/frem/Work/LUCC/data/esnet/regions/sud_gresivaudan_artif_transitions/output/"

X = np.load(folder+'X.npy')
V = np.load(folder+'V.npy')

#%%
kde = ekde.KDE(q=11,
               bounds=[
                    (0,'left'),
                    (1,'left'),
                    (2,'left'),
                   ],
               verbose=1)
kde.fit(X)

#%%
from time import time
st = time()
f = kde.predict(X)
print(time()-st)

#%%
i = 3325
print(X[i])
print(f[i])



# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt
import numpy as np
from scipy.stats import norm

x = np.array([0.1, 1.4, 1.6])
h = 0.3

y = np.linspace(-0.2, 2.6, 300)

f = np.zeros_like(y)

f_exact = norm.pdf(y, 0.0, 0.3) + 1* norm.pdf(y, 1.5, 0.3)
f_exact[y<0] = 0
f_exact /= f_exact.sum() * (y.max()-y.min()) / y.size

plt.figure(figsize=(10,3))
plt.subplot(1,2,1)

plt.plot(y, f_exact, color='tab:blue', label='exact')

for xi in x:
    j = np.abs(y-xi) < h
    f_sub = np.zeros(y.size)
    f_sub = norm.pdf(y, xi, h)
    # f_sub[y<0] = 0
    f_sub *= 1 / x.shape[0]
    f += f_sub
    if xi == x[0]:
        plt.scatter(xi, -0.04, marker='|', s=70, color='gray', label='observations')
        plt.plot(y, f_sub, color='gray', ls='--', label='Gaussian kernels')
        
    else:
        plt.plot(y, f_sub, color='gray', ls='--')
        plt.scatter(xi, -0.04, marker='|', s=70, color='gray')

plt.plot(y, f, ls='-', color='tab:red', label='kde')
plt.xlabel('z')
plt.ylabel('probability distribution')
plt.title('(a) without correction')

print(f_exact.sum() * (y.max()-y.min()) / y.size)
print(f.sum() * (y.max()-y.min()) / y.size)

plt.subplot(1,2,2)
plt.plot(y, f_exact, color='tab:blue', label='exact')
f = np.zeros_like(y)

for xi in x:
    j = np.abs(y-xi) < h
    f_sub = np.zeros(y.size)
    f_sub = norm.pdf(y, xi, h)
    f_sub[y<0] = 0
    f_sub *= 1 / x.shape[0]
    
    c = norm.cdf(xi + (xi - 0), xi, h)
    f_sub /= c
    
    f += f_sub
    if xi == x[0]:
        plt.scatter(xi, -0.04, marker='|', s=70, color='gray', label='observations')
        plt.plot(y, f_sub, color='gray', ls='--', label='Gaussian kernels')
        
    else:
        plt.plot(y, f_sub, color='gray', ls='--')
        plt.scatter(xi, -0.04, marker='|', s=70, color='gray')

plt.plot(y, f, ls='-', color='tab:red', label='kde')
plt.legend(loc='lower right', bbox_to_anchor=(1.55, 0.58))
plt.xlabel('z')
# plt.ylabel('probability distribution')
print(f.sum() * (y.max()-y.min()) / y.size)
plt.title('(b) with correction')
plt.savefig('/home/frem/Work/LUCC/gisworknotes/Articles/LUCCProbabilitiesEstimation_EnvModSoft/figures/boundary_bias.pdf', bbox_inches='tight')